<?php

namespace Drupal\field_redirect\Utils;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\Display\EntityViewDisplayInterface;
use Drupal\Core\Entity\EditorialContentEntityBase;
use Drupal\Core\Entity\EntityBase;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\file\FileInterface;
use Drupal\user\Entity\User;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\HttpException;

/**
 * Utility for Field Redirect Module.
 */
class Utility {

  use StringTranslationTrait;

  const MODULE_NAME = 'field_redirect';
  const ENTITY_PAIR_SEPARATOR = ':';
  const ENTITY_BUNDLE_WILDCARD = '*';
  const PAIR_SEPARATOR = '|';
  const FIELDS_SEPARATOR = ',';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Core Entity Storage for User Entities.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected $userEntityStorage;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The file URL generator.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected $fileUrlGenerator;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The request service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack request
   */

  /**
   * Utility constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The core 'entity_type.manager' service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The 'config.factory' service.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $file_url_generator
   *   The 'file_url_generator' service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The 'messenger' service.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   The 'request_stack' service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory,
    FileUrlGeneratorInterface $file_url_generator,
    MessengerInterface $messenger,
    RequestStack $request_stack) {
    $this->entityTypeManager = $entity_type_manager;
    $this->userEntityStorage = $this->entityTypeManager->getStorage('user');
    $this->configFactory = $config_factory;
    $this->fileUrlGenerator = $file_url_generator;
    $this->messenger = $messenger;
    $this->requestStack = $request_stack;
  }

  /**
   * Implements toString().
   *
   * @param null|array $array
   *   The mapping array.
   *
   * @return string
   *   The return value.
   */
  public function toString($array = []) {
    $return = [];
    foreach ((array) $array as $entity_type_bundle => $field_machine_names_array) {
      $field_machine_names = \implode(self::FIELDS_SEPARATOR, (array) $field_machine_names_array);
      $return[] = $entity_type_bundle . self::PAIR_SEPARATOR . $field_machine_names;
    }
    return \implode(PHP_EOL, $return);
  }

  /**
   * Implements toArray().
   *
   * @param string $mappings
   *   The mapping string.
   *
   * @return array
   *   The return value.
   */
  public function toArray($mappings = '') {
    $return = [];
    if (\strlen($mappings)) {
      if ($lines = \explode(\PHP_EOL, $mappings)) {
        foreach ($lines as $pair) {
          if (\strlen($pairs = \trim($pair))) {
            $entity_fields = \array_map('trim', \explode(self::PAIR_SEPARATOR, $pairs, 2));
            [$entity_type_bundle, $fields_str] = $entity_fields;
            $entity_type_bundle_array = \array_map('trim', \explode(self::ENTITY_PAIR_SEPARATOR, $entity_type_bundle));
            $fields_array = \array_map('trim', \explode(self::FIELDS_SEPARATOR, $fields_str));
            if (\count($entity_type_bundle_array) === 2 && !empty($fields_array)) {
              $entity_type_bundle = \implode(self::ENTITY_PAIR_SEPARATOR, $entity_type_bundle_array);
              $return[$entity_type_bundle] = $fields_array;
            }
          }
        }
      }
    }
    return $return;
  }

  /**
   * Implements isValidMappings().
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   * @param array $element
   *   The form element.
   * @param string $mappings
   *   The mapping string.
   *
   * @return bool
   *   The return value.
   */
  public function isValidMappings(FormStateInterface $form_state, array $element, $mappings = '') {
    $return = TRUE;
    if (\strlen($mappings)) {
      if ($lines = \explode(\PHP_EOL, $mappings)) {
        $line_number = 0;
        $error_t_args = [];
        $duplicates_entity = [];
        foreach ($lines as $pair) {
          if (\strlen($pairs = \trim($pair))) {
            ++$line_number;
            $pairs = \array_map('trim', \explode(self::PAIR_SEPARATOR, $pairs, 2));
            $entity_type_bundle_pair = \array_shift($pairs) ?? '';
            $field_machine_name_str = \array_shift($pairs) ?? '';
            $entity_type_bundle = \array_map('trim', \explode(self::ENTITY_PAIR_SEPARATOR, $entity_type_bundle_pair, 2));
            $entity_type = \array_shift($entity_type_bundle) ?? '';
            $entity_bundle = \array_shift($entity_type_bundle) ?? '';
            if (\strlen($entity_type) > 0 && \strlen($entity_bundle) > 0 && \strlen($field_machine_name_str) > 0) {
              if (empty($duplicates_entity[$entity_type_bundle_pair])) {
                $duplicates_entity[$entity_type_bundle_pair] = 0;
              }
              $duplicates_entity[$entity_type_bundle_pair]++;
              $entity_config_name = '';
              $entity_config_target = '';
              switch ($entity_type) {
                case 'node':
                  $entity_config_name = 'node.type.' . $entity_bundle;
                  $entity_config_target = 'type';
                  break;

                case 'taxonomy_term':
                  $entity_config_name = 'taxonomy.vocabulary.' . $entity_bundle;
                  $entity_config_target = 'vid';
                  break;

                case 'user':
                  $entity_config_name = 'user.settings';
                  $entity_config_target = 'anonymous';
                  break;

                default:
                  $error_t_args[] = $this->t('Line @line_number: Entity type "@entity_type" is not supported.', [
                    '@line_number' => $line_number,
                    '@entity_type' => $entity_type,
                  ]);
                  continue 2;
              }
              $field_config_name_prefix = '';
              $target_config_value = $this->configFactory->get($entity_config_name)->get($entity_config_target);
              if ($entity_bundle === self::ENTITY_BUNDLE_WILDCARD) {
                $field_config_name_prefix = 'field.field.' . $entity_type . '.';
              }
              elseif (FALSE
                || ($entity_type === 'user' && \strlen($target_config_value))
                || $target_config_value === $entity_bundle
              ) {
                $field_config_name_prefix = 'field.field.' . $entity_type . '.' . $entity_bundle . '.';
              }
              else {
                $error_t_args[] = $this->t('Line @line_number: The bundle name "@entity_bundle" does not exist for entity type "@entity_type".', [
                  '@line_number' => $line_number,
                  '@entity_type' => $entity_type,
                  '@entity_bundle' => $entity_bundle,
                ]);
                continue;
              }
              $field_machine_names = \array_map('trim', \explode(self::FIELDS_SEPARATOR, $field_machine_name_str));
              $duplicates_field = [];
              $max_i = \count($field_machine_names) - 1;
              foreach ($field_machine_names as $current_i => $field_machine_name) {
                if (empty($duplicates_field[$field_machine_name])) {
                  $duplicates_field[$field_machine_name] = 0;
                }
                $duplicates_field[$field_machine_name]++;
                if (\strpos($field_machine_name, '#') === 0) {
                  if ($field_machine_name !== '#403' && $field_machine_name !== '#404') {
                    $error_t_args[] = $this->t('Line @line_number: The specified status is invalid. Please specify "#403" or "#404".', [
                      '@line_number' => $line_number,
                    ]);
                  }
                  if ($max_i !== $current_i) {
                    $error_t_args[] = $this->t('Line @line_number: Status is can be specified only one at the end.', [
                      '@line_number' => $line_number,
                    ]);
                  }
                }
                else {
                  $field_config = NULL;
                  $config_factory = $this->configFactory;
                  if ($entity_bundle === self::ENTITY_BUNDLE_WILDCARD) {
                    $match_fields = $config_factory->listAll($field_config_name_prefix);
                    foreach ((array) $match_fields as $field_config_name) {
                      $config = $config_factory->get($field_config_name);
                      if (!$config->isNew() && $config->get('field_name') === $field_machine_name) {
                        $field_config = $config;
                      }
                    }
                  }
                  else {
                    $config = $config_factory->get($field_config_name_prefix . $field_machine_name);
                    if (!$config->isNew() && $config->get('field_name') === $field_machine_name) {
                      $field_config = $config;
                    }
                  }
                  if ($field_config instanceof ImmutableConfig) {
                    if (!\in_array($field_config->get('field_type'), [
                      'link',
                      'file',
                      'image',
                    ], TRUE)) {
                      $error_t_args[] = $this->t('Line @line_number: The field type for "@field_machine_name" must be link, file or image.', [
                        '@line_number' => $line_number,
                        '@field_machine_name' => $field_machine_name,
                      ]);
                    }
                  }
                  else {
                    $error_t_args[] = $this->t('Line @line_number: The field for "@field_machine_name" does not exist.', [
                      '@line_number' => $line_number,
                      '@field_machine_name' => $field_machine_name,
                    ]);
                  }
                }
              }
              foreach ($duplicates_field as $field => $number) {
                if ($number > 1) {
                  if (\strpos($field, '#') === 0) {
                    $error_t_args[] = $this->t('Line @line_number: Can specify only one status.', [
                      '@line_number' => $line_number,
                    ]);
                  }
                  else {
                    $error_t_args[] = $this->t('Line @line_number: Fields for "@field_machine_name" are specified is duplicate.', [
                      '@line_number' => $line_number,
                      '@field_machine_name' => $field_machine_name,
                    ]);
                  }
                }
              }
            }
            else {
              $error_t_args[] = $this->t('Line @line_number: The mapping format is incorrect.', [
                '@line_number' => $line_number,
              ]);
            }
          }
        }
      }
    }
    foreach ($duplicates_entity as $entity_type_bundle => $number) {
      if ($number > 1) {
        $error_t_args[] = $this->t('Set of the entity type and bundle is duplicated. (@entity_type_bundle)', [
          '@entity_type_bundle' => $entity_type_bundle,
        ]);
      }
    }
    if (!empty($error_t_args)) {
      $form_state->setError($element, Markup::create(\implode('<br>', $error_t_args)));
      $return = FALSE;
    }
    return $return;
  }

  /**
   * Implements getRedirectFields().
   *
   * @param string $target_bundle
   *   The node bundle string.
   *
   * @return array
   *   The return redirect fields.
   */
  public function getRedirectFields($target_bundle) {
    $config = $this->configFactory->get(self::MODULE_NAME . '.settings');
    $redirect_fields = $config->get(self::MODULE_NAME . '.mappings');
    return $redirect_fields[$target_bundle] ?? [];
  }

  /**
   * Implements getEntityTypeLabel().
   *
   * @param \Drupal\Core\Entity\EntityBase $entity
   *   The node bundle string.
   *
   * @return string
   *   The return entity type label.
   */
  public function getEntityTypeLabel(EntityBase $entity) {
    $target_entity_type = $entity->getEntityTypeId();
    $target_entity_bundle = $entity->bundle();

    // Get entity type config.
    $target_config_name = '';
    $target_config_key = '';
    switch ($target_entity_type) {
      case 'node':
        $target_config_name = 'node.type.' . $target_entity_bundle;
        $target_config_key = 'name';
        break;

      case 'taxonomy_term':
        $target_config_name = 'taxonomy.vocabulary.' . $target_entity_bundle;
        $target_config_key = 'name';
        break;

      case 'user':
        return $this->t('User');
    }

    // Return entity type label.
    return $this->configFactory
      ->get($target_config_name)
      ->get($target_config_key) ?? '[unknown]';
  }

  /**
   * This gets the target field and target value.
   *
   * @param array $redirect_fields
   *   The config fields with redirect rules.
   * @param \Drupal\Core\Entity\EditorialContentEntityBase|\Drupal\user\Entity\User $entity
   *   The entity in the rule.
   *
   * @return array
   *   Returns the field name and field value.
   */
  public function getTargetFields(array $redirect_fields, EntityInterface $entity) {
    // Search of target field.
    foreach ($redirect_fields as $field_name) {
      if (\strpos($field_name, '#') === 0) {
        return [$field_name, []];
      }
      elseif ($entity->hasField($field_name) && ($value = $entity->get($field_name)->getValue())) {
        return [$field_name, $value];
      }
    }
    return ['', []];
  }

  /**
   * This gets URL and a Tag from target field values.
   *
   * @param mixed $target_field_value
   *   The field with the field's value.
   *
   * @return array
   *   Returns The URL and a Tag FALSE and empty string if not found.
   */
  public function getUrlAndTag($target_field_value) {
    $url = FALSE;
    $a_tag = '';

    // If exist value of URL.
    if ($uri = $target_field_value[0]['uri'] ?? FALSE) {
      // Generate of url from $uri.
      $url = Url::fromUri($uri, ['attributes' => ['target' => '_blank']]);

      // Generate of "a" tag from $url.
      $a_tag = Link::fromTextAndUrl($url->toString(), $url)->toString();
    }

    // If exist value of target id of th file entity.
    elseif ($target_id = $target_field_value[0]['target_id'] ?? FALSE) {
      /** @var \Drupal\file\FileInterface $file */
      if (($file = $this->entityTypeManager->getStorage('file')->load($target_id)) instanceof FileInterface) {
        if ($uri = $file->get('uri')->getString()) {
          // File URL.
          $url = $this->fileUrlGenerator->generateAbsoluteString($uri);

          // Link text.
          $text = $file->getFilename();
          if (\strlen($description = $target_field_value[0]['description'] ?? '')) {
            $text .= ' (' . $description . ')';
          }

          // Generate of "a" tag from $url and $text.
          $a_tag = Markup::create('<a href="' . $url . '" target="_blank">' . $text . '</a>');
        }
      }
    }

    return [$url, $a_tag];
  }

  /**
   * This can redirect the user based on the rules in configuration.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity in the rule.
   * @param string $target_field_name
   *   The name of the field.
   * @param \Drupal\Component\Render\MarkupInterface|string $a_tag
   *   The a tag of the URL.
   */
  public function executeRedirect(EntityInterface $entity, string $target_field_name, MarkupInterface|string $a_tag) {
    $target_entity_type = $entity->getEntityTypeId();
    $target_entity_id = $entity->id();
    // Get label of entities bundle.
    $target_bundle_label = $this->getEntityTypeLabel($entity);
    // If valid entity.
    if (($entity instanceof User && $entity->isActive())
      || ($entity instanceof EditorialContentEntityBase && $entity->isPublished())
    ) {
      // Return status.
      if (\strpos($target_field_name, '#') === 0) {
        // Get name of status.
        $status = '';
        if ($target_field_name === '#403') {
          $status = '403 Forbidden';
        }
        elseif ($target_field_name === '#404') {
          $status = '404 Not Found';
        }

        // Display a status message descriptioning this
        // behavior to avoid editor confusion.
        $this->messenger->addStatus($this->t('Field Redirect Module:<br>The original behavior was that if you accessed this %target_bundle_label view directly, then return status of the @status, but because I had just edited it, was redirected to the %target_bundle_label edit form.', [
          '%target_bundle_label' => $target_bundle_label,
          '@status' => $status,
        ]));
      }

      // Redirect URL.
      else {
        // Display a status message descriptioning this
        // behavior to avoid editor confusion.
        $this->messenger->addStatus($this->t('Field Redirect Module:<br>The original behavior was that if you accessed this %target_bundle_label view directly, would be redirect to %a_tag, but since it was just edited, was redirected to the %target_bundle_label edit form.', [
          '%target_bundle_label' => $target_bundle_label,
          '%a_tag' => $a_tag,
        ]));
      }

      // For taxonomy terms, the path is slightly different.
      if ($target_entity_type === 'taxonomy_term') {
        $target_entity_type = 'taxonomy/term';
      }

      // Return to the edit form.
      $redirect_path = '/' . $target_entity_type . '/' . $target_entity_id . '/edit';
      $redirect_url = Url::fromUserInput($redirect_path)->toString();
      $redirect_response = new RedirectResponse($redirect_url);
      $redirect_response->send();
    }
  }

  /**
   * This is the logic to redirect the request.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity that was saved.
   */
  public function redirectAfterSave(EntityInterface $entity) {
    // Check instance of entity.
    if ($entity instanceof EditorialContentEntityBase || $entity instanceof User) {
      $target_entity_type = $entity->getEntityTypeId();
      $target_entity_bundle = $entity->bundle();

      // Check exists target field in entities bundle.
      $redirect_fields = [];
      $redirect_fields += $this->getRedirectFields($target_entity_type . self::ENTITY_PAIR_SEPARATOR . self::ENTITY_BUNDLE_WILDCARD);
      $redirect_fields += $this->getRedirectFields($target_entity_type . self::ENTITY_PAIR_SEPARATOR . $target_entity_bundle);

      if (empty($redirect_fields)) {
        return;
      }
      [$target_field_name, $target_field_value] = $this->getTargetFields($redirect_fields, $entity);
      if ($target_field_name === '') {
        return;
      }
      // Prepare values.
      [$url, $a_tag] = $this->getUrlAndTag($target_field_value);

      // If URL is valid, or HTTP status was setted.
      if ($url === FALSE && \strpos($target_field_name, '#') !== 0) {
        return;
      }

      $this->executeRedirect($entity, $target_field_name, $a_tag);
    }
  }

  /**
   * This may initiate the a redirect from a user visiting a view.
   *
   * @param array $build
   *   The render array.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity for the view.
   * @param \Drupal\Core\Entity\Display\EntityViewDisplayInterface $display
   *   The current display [mode] of the request.
   */
  public function redirectOnView(array &$build, EntityInterface $entity, EntityViewDisplayInterface $display) {
    // It works only when it is full page.
    if ($display->getOriginalMode() !== 'full') {
      return;
    }

    // Check instance is which.
    if ($entity instanceof EditorialContentEntityBase || $entity instanceof User) {
      $target_entity_type = $entity->getEntityTypeId();
      $target_entity_bundle = $entity->bundle();

      // Check exists target field in entities bundle.
      $redirect_fields = [];
      $redirect_fields += $this->getRedirectFields($target_entity_type . self::ENTITY_PAIR_SEPARATOR . self::ENTITY_BUNDLE_WILDCARD);
      $redirect_fields += $this->getRedirectFields($target_entity_type . self::ENTITY_PAIR_SEPARATOR . $target_entity_bundle);

      if (empty($redirect_fields)) {
        return;
      }
      [$target_field_name, $target_field_value] = $this->getTargetFields($redirect_fields, $entity);
      if ($target_field_name === '') {
        return;
      }
      // Disable the cache, because redirect must be success always.
      $build['#cache']['max-age'] = 0;

      // Prepare values.
      $redirect_url = FALSE;
      if (($entity instanceof User && $entity->isActive())
        || ($entity instanceof EditorialContentEntityBase && $entity->isPublished())
      ) {
        // If exist value of URL.
        if ($uri = $target_field_value[0]['uri'] ?? FALSE) {
          // Generate of URL from $uri.
          if (\strpos($uri, 'http://') === 0 || \strpos($uri, 'https://') === 0) {
            $redirect_url = $uri;
          }
          else {
            $redirect_url = Url::fromUri($uri)->toString();
          }
        }
        elseif ($target_id = $target_field_value[0]['target_id'] ?? FALSE) {
          /** @var \Drupal\file\FileInterface $file */
          if (($file = $this->entityTypeManager->getStorage('file')->load($target_id)) instanceof FileInterface) {
            if ($uri = $file->get('uri')->getString()) {
              // File URL.
              $redirect_url = $this->fileUrlGenerator->generateAbsoluteString($uri);
            }
          }
        }
      }

      if ($redirect_url === FALSE) {
        return;
      }
      $this->executeRedirectForView($entity, $target_field_name, $redirect_url);
    }
  }

  /**
   * This may initiate the redirect process.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The current entity for the view.
   * @param string $target_field_name
   *   The field name.
   * @param string $redirect_url
   *   The URL string to which to redirect.
   */
  public function executeRedirectForView(EntityInterface $entity, string $target_field_name, string $redirect_url) {
    // If valid entity.
    if (($entity instanceof User && $entity->isActive())
      || ($entity instanceof EditorialContentEntityBase && $entity->isPublished())
    ) {
      $target_entity_type = $entity->getEntityTypeId();
      $target_entity_id = $entity->id();

      // For taxonomy terms, the path is slightly different.
      if ($target_entity_type === 'taxonomy_term') {
        $target_entity_type = 'taxonomy/term';
      }

      // If redirect URL is valid.
      // Avoid infinite redirects.
      $request = $this->requestStack->getCurrentRequest();
      $site_url = $request->getSchemeAndHttpHost();
      $site_url__no_scheme = \preg_replace('/^[^:]+/', '', $site_url);
      $current_uri = $request->getRequestUri();
      $current_url__http = 'http' . $site_url__no_scheme . $current_uri;
      $current_url__https = 'https' . $site_url__no_scheme . $current_uri;
      $current_entity_uri = '/' . $target_entity_type . '/' . $target_entity_id;
      $current_entity_url__http = 'http' . $site_url__no_scheme . $current_entity_uri;
      $current_entity_url__https = 'https' . $site_url__no_scheme . $current_entity_uri;
      if (
        FALSE
        || $redirect_url === $current_uri
        || $redirect_url === $current_url__http
        || $redirect_url === $current_url__https
        || $redirect_url === $current_entity_uri
        || $redirect_url === $current_entity_url__http
        || $redirect_url === $current_entity_url__https
      ) {
        return;
      }

      // Do redirect to $redirect_url.
      $redirect_response = new RedirectResponse($redirect_url);
      $redirect_response->setStatusCode(307);
      $redirect_response->send();
    }
    elseif (\strpos($target_field_name, '#') === 0) {
      $status = \preg_replace('/^#/', '', $target_field_name);
      if (\ctype_digit($status)) {
        throw new HttpException($status);
      }
    }
  }

}
